import csv
import datetime
import logging
import os
import platform
import psutil
import re
import shutil
import socket
import subprocess
import sys
import tarfile
import time


#
# List of shell commands runs by Tbreport
# For this list, the parameters cannot be a variable
#       Example:
#           ethtool mgmt0       <---- mgmt0 is a variable
# To add a new command:
#       add a new line and fill the parameters
#       "command_name": ["command", "param1", "param2", "param3", ...]
#
#       "command_name" => command_name.log file will be created containing the command stdout and stderr
#       "command" => shell commmand
#       "param*" => each time there is a space required, you need to crate a new element in the list
#
#       Example:
#           'nmcli device show'
#           the nmcli command has 2 parameters => the list contain 3 elements
#           "nmcli device": ["nmcli", "device", "show"],
#
shellCommands = {
    "virt-what": ["virt-what"],
    "lscpu": ["lscpu"],
    "dmidecode system": ["dmidecode", "-t", "system"],
    "dpdk_tbtool": ["dpdk_tbtool.py", "-s"],
    "tb_exclusions.conf": ["cat", "/etc/NetworkManager/conf.d/tb_exclusions.conf"],
    "ip config": ["ifconfig", "-a"],
    "ip route": ["ip", "route", "list"],
    "ip rule": ["ip", "rule", "list"],
    "route": ["route", "-n"],
    "nmcli": ["nmcli"],
    "nmcli device": ["nmcli", "device", "show"],
    "nmcli con": ["nmcli", "con", "show"],
    "persistent-net": ["cat", "/etc/udev/rules.d/70-persistent-net.rules"],
    "journalctl": ["journalctl", "--no-pager"],
    "disk usage": ["df", "-h"],
    "memory usage": ["cat", "/proc/meminfo"],
    "ps": ["ps", "auxfww"],
    "cmdline": ["cat", "/proc/cmdline"],
    "processor info": ["cat", "/proc/cpuinfo"],
    "kernel version": ["cat", "/proc/version"],
    "os version": ["cat", "/etc/redhat-release"],
    "hwconf": ["cat", "/etc/sysconfig/hwconf"],
    "resolv": ["cat", "/etc/resolv.conf"],
    "crontab": ["crontab", "-l"],
    "lstopo": ["lstopo-no-graphics"],
    "vmware cpures": ["vmware-toolbox-cmd", "stat", "cpures"],
    "vmware memres": ["vmware-toolbox-cmd", "stat", "memres"],
}

TIMESTAMP_FORMAT = "%Y-%m-%d-%H-%M-%S"
CTIME_TIME_FORMAT = "%a %b %d %H:%M:%S %Y"
DATE_FORMAT = "%Y-%m-%d"

REGEX_DAY_STRICT = r"^\d{4}-\d{2}-\d{2}$"
REGEX_FILE_DAY = r".*(?P<date>\d{4}-\d{2}-\d{2}).*"
REGEX_FILE_DAY_AND_HOUR = r".*(?P<date>\d{4}-\d{2}-\d{2})(?P<hour>_\d{2}-\d{2}-\d{2})?.*"
REGEX_FILE_DAY_DATE = "date"
REGEX_SYSTEM_ID = r"^\d{5}$"
REGEX_MAJOR_DIRECTORY = r"^\d{1}\.\d{1,2}$"
REGEX_PACKAGE_DIRECTORY = r"^\d{1}\.\d{1,2}\.\d{1,3}\.{0,1}\d{0,2}$"
REGEX_VAR_LOG_MESSAGES = r"^messages.*"
REGEX_LEASE = r".*\.lease$"
REGEX_IF_CFG = r"^ifcfg-.*"
REGEX_LOG = r".*\.log$"
REGEX_MY_CNF = r"my\.cnf"
REGEX_RESOLV_CONF = r"resolv\.conf"
REGEX_TOOLPACK_DB_NAME = r"^toolpack_(?P<dbName>.*)"
REGEX_TOOLPACK_SERVICE_APP = r"^application= (?P<wdPath>.*)"
REGEX_TOOLPACK_SERVICE_APP_WDPATH = "wdPath"
REGEX_TOOLPACK_WD_PKG_PATH = r"^TB_PKG_PATH.*= (?P<pkgPath>.*)"
REGEX_TOOLPACK_WD_PKG_PATH_PKGPATH = "pkgPath"
REGEX_TOOLPACK_BIN_PATH = r"^APPLICATION_PATH.*= @{TB_PKG_PATH}(?P<appPath>.*)/tboamapp"
REGEX_TOOLPACK_BIN_PATH_APPPATH = "appPath"
REGEX_TELNETDUMP_FILE = r"^tbtelnetdump_.*.dmp$"
REGEX_PCAP_FOLDER = r"^(?P<date>\d{4}-\d{2}-\d{2})_\d{2}-\d{2}-\d{2}$"
REGEX_PCAP_FOLDER_DATE = "date"

TOOLPACK = "toolpack"
TOOLPACK_SETUP = "setup"
TOOLPACK_PKG = "pkg"
TOOLPACK_STATUS = "status"
LICENSES_DIRECTORY = "licenses"
TBROUTER_DIR = "tbrouter"
WEB_DIR = "web"
DUMPS_DIR = "dumps"
SYSTEM_DIR = "system"
SYSCONFIG_DIR = "sysconfig"
NETWORK_SCRIPTS_DIR = "network-scripts"
CONFIG_DIR = "config"
LOGS_DIR = "logs"
LOG_DIR = "log"
LIB_DIR = "lib"
DHCLIENT_DIR = "dhclient"
NETWORK_MANAGER_DIR = "NetworkManager"
MARIADB_DIR = "mariadb"
STATUS_DIR = "status"
VAR_DIR = "/var"
ETC_DIR = "/etc"
EXTENTION_LIST = [".log", ".log.gz", ".uct", ".uct.gz", ".stath.gz", ".stath"]
SKIPPED_EXTENTION_LIST = [".tar.gz"]
TBREPORT_EXTENSION = ".tar.gz"
DUMPS_FILENAME = "dumps_{}.csv"
CORE_FILE = "core."
CORE_FILENAME = "core.csv"
CORE_DUMP_HEADER_ROW = ["filename", "path", "creationtime"]
TBREPORT_NAME = "tbreport_{}"
TBREPORT_LOG_FILENAME = "tbreport{}.log"
TBSTATUS_LOG_FILE = "tbstatus.log"
LINUX_BASE_DIR = "/lib/tb"
WINDOWS_BASE_DIR = "C:\\TelcoBridges"
SERVICE_CONFIG_DIR = "/lib/tb/toolpack/setup/{}/service/config.txt"
ERROR = "Error: {}"


class TBPath:
    """
    Class that lists toolpack system paths used by tbreport.
    """

    def __init__(self, systemId):

        if platform.system() == "Linux":
            logging.info("Linux system")
            self.platform = "Linux"
            self.baseDir = os.path.join(LINUX_BASE_DIR)
        else:
            logging.info("Windows system")
            self.platform = "Windows"
            self.baseDir = os.path.join(WINDOWS_BASE_DIR)

        self.rootDir = os.path.join(self.baseDir, TOOLPACK)
        self.setupDir = os.path.join(self.rootDir, TOOLPACK_SETUP)
        self.pkgDir = os.path.join(self.rootDir, TOOLPACK_PKG)
        self.systemPath = os.path.join(self.setupDir, systemId)
        self.serviceConfig = SERVICE_CONFIG_DIR.format(systemId)
        

class TBFileInfo:
    """
    Class that keeps the name, path and creation time of a file.
    """

    def __init__(self, name, path):
        self.name = name
        self.path = path

        fileFullPath = os.path.join(self.path, self.name)
        cTime = os.path.getctime(fileFullPath)
        self.creationTime = datetime.datetime.strptime(time.ctime(cTime), CTIME_TIME_FORMAT)


class TBCore:
    """
    Class that creates the .csv file containing the creation time of every core.
    """

    def __init__(self):

        self.files = []

    def createCoreInfoFile(self, filePath):
        """
        Create the core.csv file to the specified path.

        _filePath_      file directory where it will be created.
        """

        fullpath = os.path.join(filePath, CORE_FILENAME)
        logging.info("Creating {} in {}".format(CORE_FILENAME, filePath))
        try:
            with open(fullpath, 'w') as csvfile:
                fileWriter = csv.writer(csvfile, delimiter=",", lineterminator="\n")
                fileWriter.writerow(CORE_DUMP_HEADER_ROW)
                # Add 1 row per core file
                for coreFile in self.files:
                    creationTime = coreFile.creationTime.strftime(TIMESTAMP_FORMAT)
                    row = [coreFile.name, coreFile.path, creationTime]
                    fileWriter.writerow(row)

        except Exception as e:
            logging.error(ERROR.format(e.message))
            logging.error("Failed to create core file: {}".format(fullpath))


class TBDumps:
    """
    Class that keeps all the dumps file and creates the .csv file.
    """

    def __init__(self):

        self.dstDir = None
        self.appName = None
        self.files = []

    def createStatsFile(self):
        """
        Create a .csv file containing the creation time of dumps file.
        """

        filename = DUMPS_FILENAME.format(self.appName)
        fullpath = os.path.join(self.dstDir, filename)
        logging.info("Creating {} in {}".format(filename, self.dstDir))
        try:
            with open(fullpath, 'w') as csvfile:
                fileWriter = csv.writer(csvfile, delimiter=",", lineterminator="\n")
                fileWriter.writerow(CORE_DUMP_HEADER_ROW)
                # read file row by row
                for coreFile in self.files:
                    creationTime = coreFile.creationTime.strftime(TIMESTAMP_FORMAT)
                    row = [coreFile.name, coreFile.path, creationTime]
                    fileWriter.writerow(row)

        except Exception as e:
            logging.error(ERROR.format(e.message))
            logging.error("Failed to create dumps file: {}".format(fullpath))

    def addDumpFile(self, srcFile, dstFile):
        """
        Adds a dump file directory and its time creation in the files attribute from this class.

        _srcFile_       source of the file.
        _dstFile_       destination of the file.
        """

        srcDir, srcFilename = os.path.split(srcFile)
        appDir, dumpDir = os.path.split(srcDir)
        if dumpDir == DUMPS_DIR:
            # get application name where dmp files are located
            _, appName = os.path.split(appDir)
            self.dstDir = dstFile
            self.appName = appName
            self.files.append(TBFileInfo(srcFilename, srcDir))


class TBMajor:
    """
    Class that contains a major version and its path
    """

    def __init__(self, path, version):
        self.version = version
        self.path = path


class TBSystem:
    """
    Class that keeps specific information of the system like its id, majors, corefiles and toolpack path
    """

    def __init__(self, systemId):

        if type(systemId) is not str:
            systemId = str(systemId)
        self.systemId = systemId
        self.majors = []
        self.toolpack = TBPath(self.systemId)
        self.__getMajors()
        self.coreFiles = {}
        
    def __getMajors(self):
        """
        Retrieve all major versions path of toolpack.
        """

        logging.info("Retrieving majors: {}".format(self.toolpack.systemPath))
        subFiles = os.listdir(self.toolpack.systemPath)

        # get all directories' name and checks if they fit the major directory regex
        for subFile in subFiles:
            logging.debug("subFile {}".format(subFile))
            subFileFullPath = os.path.join(self.toolpack.systemPath, subFile)
            if os.path.isdir(subFileFullPath):
                mSubDir = re.search(REGEX_MAJOR_DIRECTORY, subFile)
                if mSubDir:
                    logging.debug("major found {}".format(subFile))
                    self.majors.append(TBMajor(subFileFullPath, subFile))


class TBReportPath:
    """
    Class that keeps the tbreport, pkg and setup directory.
    """

    def __init__(self, tbreportName):

        self.rootDir = os.path.join(os.getcwd(), tbreportName, socket.gethostname())
        self.pkgDir = os.path.join(self.rootDir, TOOLPACK_PKG)
        self.setupDir = os.path.join(self.rootDir, TOOLPACK_SETUP)


class TBReport:
    """
    Class that retrieves and copy all the relevant files for the TBReport.
    """

    def __init__(self, logFile, time, firstDay, lastDay=None, systemId="12358"):

        logging.info("Initialize report")
        self.logFile = logFile
        self.__validateDateRange(firstDay, lastDay)
        self.os = os.name
        self.platform = platform.system()
        self.systemPath = TBPath(systemId)
        self.system = TBSystem(systemId)
        self.time = time
        self.tbreportName = TBREPORT_NAME.format(self.time)
        self.hostname = socket.gethostname()
        self.reportPath = TBReportPath(self.tbreportName)
        
        logging.debug("Directory where files are copied to (working directory): {}".format(self.reportPath.rootDir))
        logging.debug("Directory where files are copied from : {}".format(self.systemPath.rootDir))
        if (self.reportPath.rootDir.find(self.systemPath.rootDir) != -1):
            raise ValueError("Can't run TBReport.py in Toolpack system directory. Please change current working directory,\nso it is not inside {}".format(self.systemPath.rootDir))
        
        self.diskUsage = psutil.disk_usage(os.getcwd())
        logging.debug("Total space on disk: {}".format(self.diskUsage.total))
        logging.debug("Used space on disk: {}".format(self.diskUsage.used))

        if (self.diskUsage.used / self.diskUsage.total >= 0.95):
            raise RuntimeError("Can't run TBReport.py if there is less than 5% space left on disk")

        if self.diskUsage.free / (1024.0 ** 3) < 1:
            raise RuntimeError("Can't run TBReport.py if there is less than 1GB of space left on disk")

        self.coreDumps = TBCore()

        logging.info("System Info:")
        logging.info("OS: {}".format(self.os))
        logging.info("Platform: {}".format(self.platform))
        logging.info("Creating report for {} to {}".format(self.firstDay, self.lastDay))

        if platform.system() == "Linux":
            process = psutil.Process(os.getpid())
            # set priority to low (-20 highest, 19 lowest)
            process.nice(10)

    def __validateDateRange(self, firstDay, lastDay):
        """
        Validates if the first day and last day are in a valid format (format: yyyy-mm-dd) and set the last day
        if the user didn't specify one.

        _firstDay_      first day of the date range.
        _lastDay_       last day of the date range.
        """
        # validate first day.
        self.firstDay = firstDay
        self.firstDayTimestamp = self.__validateDate(firstDay)

        # validate last day.
        if lastDay is None:
            self.lastDay = firstDay
            self.lastDayTimestamp = self.firstDayTimestamp + datetime.timedelta(days=1)
        else:
            self.lastDay = lastDay
            self.lastDayTimestamp = self.__validateDate(lastDay)
            self.lastDayTimestamp = self.lastDayTimestamp + datetime.timedelta(days=1)

    def __validateDate(self, day):
        """
        Validates if the date is in a valid format (yyyy-mm-dd).

        _day_       date to validate in string.

        Returns the date in a datetime object if it is valid.
        """
        
        mDate = re.match(REGEX_FILE_DAY_AND_HOUR, day)
        if mDate:
            timestamp = datetime.datetime.strptime(mDate.group('date'), DATE_FORMAT)
        else:
            raise ValueError("Invalid first day format receive: {}, expecting: yyyy-mm-dd".format(day))

        return timestamp

    def createSystemReport(self):
        """
        Creates the system report and compress it.
        """

        logging.info("Creating temp directory")
        try:
            # Get toolpack info
            self.getToolpackCurrentConfig()

            # Get setup toolpack logs
            self.getToolpackSetupLogs()

            # Get pkg toolpack logs
            self.getToolpackPkgLogs()

            # Get system information
            self.getSystemInformation()

            # Get toolpack status
            self.getSystemStatus()
            
            #
            # Warning: do not add new files in the tbreport beyond this point.
            # The tbreport log file and compression process is starting
            #

        except Exception as e:
            logging.error("tbreport errors: {}".format(e.message))

        finally:
            self.compressReport()

    def compressReport(self):
        """
        Compress the system report in a tar.gz file
        """

        # Copy log file in tmpdir
        logging.info("Copying tbreport log file in report directory")
        dstLogFile = os.path.join(self.reportPath.rootDir, self.logFile)
        srcLogFile = os.path.join(os.getcwd(), self.logFile)
        shutil.copyfile(srcLogFile, dstLogFile)

        # gzip tmp dir
        logging.info("Compress tbreport")
        self.tbreportZipName = self.tbreportName + TBREPORT_EXTENSION
        with tarfile.TarFile.gzopen(self.tbreportZipName, mode="w") as zipReport:
            zipReport.add(self.tbreportName)

        # delete tmp dir
        tmpdir = os.path.join(os.getcwd(), self.tbreportName)
        logging.info("Deleting {}".format(tmpdir))
        shutil.rmtree(tmpdir)

    def needsFiles(self, fileName, filePath, keepFilesWithDateOnly=True):
        """
        Checks if the file is a relevant (have an extension from EXTENTION_LIST) file to the report.

        _fileName_                  name of the file.
        _filePath_                  path of the file.
        _keepFilesWithDateOnly_     only keep files with extension in EXTENSION_LIST

        Return true is the file is relevant.
        """

        if fileName.startswith(CORE_FILE):
            coreInfo = TBFileInfo(fileName, filePath)
            self.coreDumps.files.append(coreInfo)
            return False

        if fileName.endswith(tuple(EXTENTION_LIST)):
            fileWithDate = re.search(REGEX_FILE_DAY, fileName)
            if fileWithDate:
                # Checks if directory is in the date range
                fileTime = datetime.datetime.strptime(fileWithDate.group(REGEX_FILE_DAY_DATE), DATE_FORMAT)
                if fileTime < self.firstDayTimestamp or fileTime >= self.lastDayTimestamp:
                    return False
        else:
            if keepFilesWithDateOnly:
                return False
            if fileName.endswith(tuple(SKIPPED_EXTENTION_LIST)):
                return False

        logging.debug("Filename: {}".format(fileName))
        return True

    def copyNeedsFiles(self, root, subFiles, subPathFolder, destFile):
        """
        Copies the dump files to the specified destination and create a TBDumps object with those dump files.

        _root_              root of the source path.
        _subFiles_          all the files in the root path.
        _subpathFolder_     start of the path to get the relative path of the source file.
        _destFile_          destination of the files.
        """

        tbDumps = TBDumps()

        # Go through every subfiles checking for dump files
        for subFile in subFiles:
            if self.needsFiles(subFile, root, False):
                sourceFile = os.path.join(root, subFile)
                subPath = os.path.relpath(sourceFile, subPathFolder)
                destinationFile = os.path.join(destFile, subPath)
                logging.info("Copying {} => {}".format(sourceFile, destinationFile))
                destinationFileDir = os.path.dirname(destinationFile)

                # Checks is the directory already exists
                if not os.path.exists(os.path.dirname(destinationFile)):
                    logging.info("Creating subdirectory {}".format(destinationFileDir))
                    os.makedirs(destinationFileDir)

                tbDumps.addDumpFile(sourceFile, destinationFileDir)
                shutil.copyfile(sourceFile, destinationFile)

        if tbDumps.files:
            tbDumps.createStatsFile()

    def getToolpackCurrentConfig(self):
        """
        Gets logs from the Toolpack config.
        """

        if platform.system() == "Linux":
            # Looking for the wdPath from the serviceConfig file
            with open(self.systemPath.serviceConfig, 'r') as serviceFile:
                # Reading every line in reversed order
                for serviceLine in reversed(serviceFile.readlines()):
                    logging.debug("serviceLine => {}".format(serviceLine))
                    # Checking for the application line
                    oamappWdCfgRe = re.search(REGEX_TOOLPACK_SERVICE_APP, serviceLine)
                    if oamappWdCfgRe:
                        with open(oamappWdCfgRe.group(REGEX_TOOLPACK_SERVICE_APP_WDPATH)) as wdFile:
                            # Looking for the pkg path and the application path
                            for wdLine in wdFile.readlines():
                                logging.debug("wdLine => {}".format(wdLine))
                                pkgPathRe = re.search(REGEX_TOOLPACK_WD_PKG_PATH, wdLine)
                                if pkgPathRe:
                                    pkgPath = pkgPathRe.group(REGEX_TOOLPACK_WD_PKG_PATH_PKGPATH)

                                appPathRe = re.search(REGEX_TOOLPACK_BIN_PATH, wdLine)
                                if appPathRe:
                                    appPath = appPathRe.group(REGEX_TOOLPACK_BIN_PATH_APPPATH)
                        
                        self.tbbin = pkgPath + appPath
                        logging.info("tbbin => '{}'".format(self.tbbin))
                        break

    def getToolpackSetupLogs(self):
        """
        Gets logs from the Toolpack setup.
        """

        logging.info("Gathering Toolpack setup logs")
        os.makedirs(self.reportPath.setupDir)
        
        # Copy all the major directories
        for major in self.system.majors:
            reportMajorDir = os.path.join(self.reportPath.setupDir, major.version)
            logging.info("reportMajorDir: {}".format(reportMajorDir))
            os.mkdir(reportMajorDir)

            for root, _, subFiles in os.walk(major.path):
                self.copyNeedsFiles(root, subFiles, major.path, reportMajorDir)

        # get tbrouter logs
        self.getTbrouterLogs()

        # send core file information to file
        self.coreDumps.createCoreInfoFile(self.reportPath.setupDir)

    def getTbrouterLogs(self):
        """
        Gets logs from TBRouter.
        """

        tbrouterDir = os.path.join(self.systemPath.rootDir, TBROUTER_DIR)
        tbrouterTargetDir = os.path.join(self.reportPath.rootDir, TOOLPACK_SETUP, TBROUTER_DIR)
        os.makedirs(tbrouterTargetDir)

        # Go through every directories and subfiles recursively
        for root, subDirs, subFiles in os.walk(tbrouterDir):
            logging.debug("Current Root: {}".format(root))
            pcapFolder = re.search(REGEX_PCAP_FOLDER, os.path.basename(root))
            if pcapFolder:
                # Checks if directory is in the date range
                fileTime = datetime.datetime.strptime(pcapFolder.group(REGEX_PCAP_FOLDER_DATE), DATE_FORMAT)
                if fileTime < self.firstDayTimestamp or fileTime >= self.lastDayTimestamp:
                    # empty subDirs and subFiles so we don't unnecessary go in the out of range directories
                    subDirs[:] = []
                    subFiles[:] = []

            self.copyNeedsFiles(root, subFiles, tbrouterDir, tbrouterTargetDir)

    def getToolpackPkgLogs(self):
        """
        Gets logs from Toolpack pkg.
        """

        logging.info("Gathering Toolpack pkg logs")
        os.makedirs(self.reportPath.pkgDir)

        # Go through every directories and subfiles recursively
        for root, subDirs, _ in os.walk(self.system.toolpack.pkgDir):
            logging.info("directory:{}".format(root))
            for subDir in subDirs:
                logging.debug("subdir {}".format(subDir))
                isPackageDir = self.getPackageLogs(subDir)
                if not isPackageDir:
                    self.getLicenses(root, subDir)

            break

    def getPackageLogs(self, directory):
        """
        Gets logs from a pkg directory.

        Returns true if it was a pkg directory.
        """

        package = re.search(REGEX_PACKAGE_DIRECTORY, directory)
        if not package:
            logging.debug("{} not a package directory: ".format(directory))
            return False

        logging.info("Package found {}".format(directory))

        webDir = os.path.join(self.system.toolpack.pkgDir, directory, WEB_DIR)

        # Go through every directories and subfiles recursively
        for root, _, subFiles in os.walk(webDir):
            for subFile in subFiles:
                if self.needsFiles(subFile, root):
                    sourceFile = os.path.join(root, subFile)
                    destinationDir = os.path.join(self.reportPath.pkgDir, directory, WEB_DIR)
                    if not os.path.exists(os.path.dirname(destinationDir)):
                        os.makedirs(destinationDir)
                    destinationFile = os.path.join(destinationDir, subFile)
                    shutil.copyfile(sourceFile, destinationFile)

        return True

    def getSystemInformation(self):
        """
        Gets information about the system.
        """

        self.reportSystemDir = os.path.join(self.reportPath.rootDir, SYSTEM_DIR)
        self.reportSystemStatusDir = os.path.join(self.reportSystemDir, STATUS_DIR)
        os.makedirs(self.reportSystemStatusDir)

        self.reportSystemLogsDir = os.path.join(self.reportSystemDir, LOGS_DIR)
        os.makedirs(self.reportSystemLogsDir)

        self.reportSystemConfigDir = os.path.join(self.reportSystemDir, CONFIG_DIR)
        os.makedirs(self.reportSystemConfigDir)

        self.runShellCommands()
        self.getSystemLogs()
        self.getSystemConfig()

        self.getSystemToolpackDatabases()

    def getSystemLogs(self):
        """
        Gets logs from the system logs.
        """

        if self.platform != "Linux":
            return
        logging.info("Gathering system logs")

        varLogDir = os.path.join(VAR_DIR, LOG_DIR)
        self.copySystemLogs(varLogDir, self.reportSystemLogsDir, REGEX_VAR_LOG_MESSAGES)

        dhclientDir = os.path.join(VAR_DIR, LIB_DIR, DHCLIENT_DIR)
        self.copySystemLogs(dhclientDir, self.reportSystemLogsDir, REGEX_LEASE)

        netWorkManagerDir = os.path.join(VAR_DIR, LIB_DIR, NETWORK_MANAGER_DIR)
        self.copySystemLogs(netWorkManagerDir, self.reportSystemLogsDir, REGEX_LEASE)

        mariaDbDir = os.path.join(VAR_DIR, LOG_DIR, MARIADB_DIR)
        self.copySystemLogs(mariaDbDir, self.reportSystemLogsDir, REGEX_LOG)

    def getSystemStatus(self):
        """
        Gets logs from the system status.
        """

        if self.platform != "Linux":
            return
            
        self.reportStatusDir = os.path.join(self.reportPath.rootDir, TOOLPACK_STATUS)
        os.makedirs(self.reportStatusDir)

        logging.info("Running tb640debug")
        command = ["tb640debug", "-h", "-ALL"]
        commandOutput = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=self.reportStatusDir)

        logging.info("Running tbtelnetdump")
        command = ["./tbtelnetdump", "-adapter", "all", "-c", "../../../tbtelnetdump/config.txt"]
        commandOutput = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=self.tbbin)
        # copy telnet dump files
        self.copySystemLogs(self.tbbin, self.reportStatusDir, REGEX_TELNETDUMP_FILE, move=True)

        logging.info("Running tbstatus")
        command = ["tbstatus", "/*"]
        commandOutput = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=self.reportStatusDir)
        tbstatusOutput, _ = commandOutput.communicate()
        logging.debug("tbstatus out:{}".format(tbstatusOutput))
        statusLogFile = os.path.join(self.reportStatusDir, TBSTATUS_LOG_FILE)
        with open(statusLogFile, 'w') as statusFile:
            statusFile.write(tbstatusOutput)

    def getSystemConfig(self):
        """
        Get logs from the system config.
        """

        if self.platform != "Linux":
            return
        logging.info("Gathering system config")
        
        networkScriptDir = os.path.join(ETC_DIR, SYSCONFIG_DIR, NETWORK_SCRIPTS_DIR)
        self.copySystemLogs(networkScriptDir, self.reportSystemConfigDir, REGEX_IF_CFG)

        mariaDbDir = os.path.join(ETC_DIR)
        self.copySystemLogs(mariaDbDir, self.reportSystemConfigDir, REGEX_MY_CNF)

        resolvDir = os.path.join(ETC_DIR)
        self.copySystemLogs(resolvDir, self.reportSystemConfigDir, REGEX_RESOLV_CONF)
        
    def copySystemLogs(self, srcDir, dstDir, regex, recursive=False, move=False):
        """
        Copies the system logs if they match the regex passed in the paramters.

        _srcDir_        source directory.
        _dstDir_        destination directory.
        _regex_         regex to match.
        _recursive_     go through folders and files recursively?
        _move_          move the file? (false => copy the file instead of moving it).
        """

        for root, _, subFiles in os.walk(srcDir):
            for subFile in subFiles:
                regMatch = re.search(regex, subFile)
                if regMatch:
                    sourceFile = os.path.join(root, subFile)
                    destinationFile = os.path.join(dstDir, subFile)
                    logging.info("Copying file {} => {}".format(sourceFile, destinationFile))
                    if move:
                        shutil.move(sourceFile, destinationFile)
                    else:
                        shutil.copyfile(sourceFile, destinationFile)
            
            if not recursive:
                break

    def getSystemToolpackDatabases(self):
        """
        Creates a database backup for every database from the Toolpack.
        """

        # list all toolpack db
        # mysql -utbdb -ptbdbpw -e "show databases;"
        command = ["mysql",
                   "-utbdb",
                   "-ptbdbpw",
                   "-e",
                   "show databases;"]
        
        commandOutput = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, _ = commandOutput.communicate()

        databases = stdout.splitlines()
        for database in databases:
            dbName = re.search(REGEX_TOOLPACK_DB_NAME, database)
            if dbName:
                logging.info("database found: {}".format(database))
                # create database command
                dbBackupCommand = ["mysqldump",
                                   "--host={}".format(self.hostname),
                                   "-utbdb",
                                   "-ptbdbpw",
                                   "--skip-add-locks",
                                   "--skip-lock-tables",
                                   "--skip-quick",
                                   "--single-transaction",
                                   "--opt",
                                   database]
                commandOutput = subprocess.Popen(dbBackupCommand, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                stdout, _ = commandOutput.communicate()

                databaseBackup = "{}/{}.sql".format(self.reportPath.setupDir, database)
                logging.info("Creating database dump in {}".format(databaseBackup))
                with open("{}".format(databaseBackup), 'w') as sqlBackup:
                    sqlBackup.write(stdout)

    def runShellCommands(self):
        """
        Runs every shell commands from the constant shellCommands and prints each output in a different file.
        """

        if self.platform != "Linux":
            return

        for name, command in shellCommands.iteritems():

            try:
                logging.info("running command:'{}'".format(" ".join(command)))
                commandOutput = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                stdout, stderr = commandOutput.communicate()
            except Exception as e:
                logging.error("Failed to run command: {}".format(command))
                logging.error(ERROR.format(e.message))

            # Putting the output of the command in a folder
            commandFileName = name.replace(" ", "_") + ".log"
            fullFileName = os.path.join(self.reportSystemStatusDir, commandFileName)
            with open(fullFileName, 'w') as commandFile:
                if stdout:
                    commandFile.write("stdout:\n")
                    commandFile.write(stdout)
                if stderr:
                    commandFile.write("stderr:\n")
                    commandFile.write(stderr)

    def getLicenses(self, parentDir, directory):
        """
        Copies the license folder if the directory is a Licenses directory.

        _parentDir_     parent directory.
        _directory_     name of the folder to check.

        Returns true if the directory was a license directory.
        """

        # Checks if the directory is a license directory
        licenses = re.search(LICENSES_DIRECTORY, directory)
        if not licenses:
            logging.info("{} not a license directory".format(directory))
            return False

        logging.info("License directory found!")
        licenseDir = os.path.join(parentDir, directory)
        dstLicenseDir = os.path.join(self.reportPath.pkgDir, LICENSES_DIRECTORY)
        os.makedirs(dstLicenseDir)
        self.copySystemLogs(licenseDir, dstLicenseDir, r".*.dat")
        return True

    def returnZipFileName(self):
        """
        Returns the name of the zip file containing all log and configs
        """
        return os.path.join(os.getcwd(), self.tbreportZipName)


def printUsage():
    """
    Prints how to use this script.
    """

    logging.info("********************************************")
    logging.info("* Usage:")
    logging.info("    Script is expecting up to 3 optional arguments")
    logging.info("")
    logging.info("      1- First day to gather (None = today)")
    logging.info("      2- Last day to gather (None = First day)")
    logging.info("      3- system ID (None = 12358")
    logging.info("")
    logging.info("Examples:")
    logging.info("  Getting today's logs:")
    logging.info("    python TBReport.py")
    logging.info("")
    logging.info("  Getting logs between 2020-02-27 and 2020-02-28")
    logging.info("    python TBReport.py 2020-02-27 2020-02-28")
    logging.info("")
    logging.info("********************************************")


if __name__ == '__main__':

    loggingLevel = logging.DEBUG
    # loggingLevel = logging.DEBUG
    rootLogger = logging.getLogger("")
    rootLogger.setLevel(loggingLevel)
    # set log format
    formatter = logging.Formatter("%(asctime)s %(levelname)s %(message)s",
                                  datefmt=TIMESTAMP_FORMAT)
    timeFormat = datetime.datetime.now().strftime(TIMESTAMP_FORMAT)
    # Prepare file log
    logFile = TBREPORT_LOG_FILENAME.format(timeFormat)
    fileHandler = logging.FileHandler(logFile, 'w')
    
    fileHandler.setFormatter(formatter)
    # Prepare console log
    streamHandler = logging.StreamHandler()
    streamHandler.setFormatter(formatter)
    # log in file and console
    rootLogger.addHandler(fileHandler)
    rootLogger.addHandler(streamHandler)

    logging.info("sys.argv:{} => {}".format(sys.argv, len(sys.argv)))

    firstDate = None
    lastDate = None
    systemId = "12358"
    if len(sys.argv) > 1:
        if sys.argv[1] == "--help" or sys.argv[1] == "-h":
            printUsage()
            sys.exit(1)
        firstDate = sys.argv[1]
    if len(sys.argv) > 2:
        lastDate = sys.argv[2]
    if len(sys.argv) > 3:
        systemId = sys.argv[3]

    if not firstDate:
        today = datetime.date.today()
        firstDate = today.strftime(DATE_FORMAT)

    report = TBReport(logFile, timeFormat, firstDate, lastDate, systemId)
    report.createSystemReport()
    print("Path to output gzipped file of TBreport.py, to be used by TBDebug client:")
    # Important: Never remove line under here. TBDebug won't be able to find where the ouptut file of TBReport.py is. Thank you
    print(report.returnZipFileName())
    # Important: Never remove this last function
